/*-
 * %%----------------------------------------------------------------------------------------------
 * Solidify Framework - Solidify Util - ZipTool.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.unige.solidify.util;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Collections;
import java.util.Enumeration;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ch.unige.solidify.SolidifyConstants;

public class ZipTool {
  private static final Logger log = LoggerFactory.getLogger(ZipTool.class);

  public static boolean unzipFiles(InputStream zipData, Path destination, List<String> list, String relativeFolder) {
    int count = 0;
    try (ZipInputStream zis = new ZipInputStream(zipData)) {
      ZipEntry ze = zis.getNextEntry();
      while (ze != null) {
        if (isFileCandidate(list, ze)) {
          extractFile(zis, ze, destination, relativeFolder);
          count++;
        }
        ze = zis.getNextEntry();
      }
      zis.closeEntry();

    } catch (final IOException e) {
      log.error("Problem in unzipping package to {}", destination, e);
      return false;
    }
    log.info("Unzip package in '{}': {} entries extracted", destination, count);
    return true;

  }

  private static Path checkDestination(ZipEntry zipEntry, Path destination, String relativeFolder)
          throws IOException {
    Path newFile;
    if (relativeFolder == null) {
      newFile = destination.resolve(zipEntry.getName()).toAbsolutePath();
    } else {
      final String newName = zipEntry.getName()
              .substring(zipEntry.getName().indexOf(normalizePath(relativeFolder)) + relativeFolder.length());
      newFile = destination.resolve(newName).toAbsolutePath();
    }
    if (zipEntry.isDirectory()) {
      FileTool.ensureFolderExists(newFile);
      return null;
    } else {
      if (!FileTool.ensureFolderExists(newFile.getParent())) {
        final String message = "Cannot create parent directory (" + newFile + ")";
        log.error(message);
        throw new IOException(message);
      }
      return newFile;
    }
  }

  private static void extractFile(ZipInputStream zipFile, ZipEntry zipEntry, Path destination, String relativeFolder) throws IOException {
    final Path newFile = checkDestination(zipEntry, destination, relativeFolder);
    if (newFile != null) {
      try (BufferedOutputStream bos = new BufferedOutputStream(new FileOutputStream(newFile.toFile()), SolidifyConstants.BUFFER_SIZE)) {
        final byte[] buffer = new byte[SolidifyConstants.BUFFER_SIZE];
        int len;
        while ((len = zipFile.read(buffer)) > 0) {
          bos.write(buffer, 0, len);
        }
      }
    }
  }

  private static boolean isFileCandidate(List<String> list, ZipEntry ze) {
    if (list == null) {
      return true;
    }
    for (final String f : list) {
      final String zePath = ze.getName();
      if (zePath.startsWith(normalizePath(f))) {
        return true;
      }
    }
    return false;
  }

  private static String normalizePath(String path) {
    return path.replace("\\", "/");
  }

  private final Path archive;

  public ZipTool(String archive) {
    this.archive = Paths.get(archive);
  }

  public ZipTool(URI archive) {
    this.archive = Paths.get(archive);
  }

  public boolean addFilesToZip(Path source, List<Path> files, List<Path> destinationPath, boolean excludeContainingFolder) {
    files = files.stream().map(source::relativize).collect(Collectors.toList());
    return this.createZip(source, files, destinationPath, excludeContainingFolder);
  }

  public boolean checkZip() {
    try (ZipFile zipFile = new ZipFile(this.getArchive())) {
      final Enumeration<? extends ZipEntry> entries = zipFile.entries();
      while (entries.hasMoreElements()) {
        final ZipEntry zipEntry = entries.nextElement();
        zipFile.getInputStream(zipEntry);
        zipEntry.getCrc();
        zipEntry.getCompressedSize();
        zipEntry.getName();
      }
      return true;
    } catch (final IOException e) {
      log.error("IOException when checking zip file {}", this.getArchive(), e);
      return false;
    }
  }

  public boolean contains(String fileName) {
    try (ZipInputStream zis = new ZipInputStream(
            new BufferedInputStream(new FileInputStream(this.getArchive()), SolidifyConstants.BUFFER_SIZE))) {
      ZipEntry entry;
      while ((entry = zis.getNextEntry()) != null) {
        if (entry.getName().equals(fileName)) {
          return true;
        }
      }
    } catch (final IOException e) {
      log.error("Problem in reading package ({})", this.archive, e);
    }
    return false;
  }

  public boolean containsFolder(String folderName) {
    try (ZipInputStream zis = new ZipInputStream(
            new BufferedInputStream(new FileInputStream(this.getArchive()), SolidifyConstants.BUFFER_SIZE))) {
      ZipEntry entry;
      while ((entry = zis.getNextEntry()) != null) {
        if (entry.getName().startsWith(folderName)) {
          return true;
        }
      }
    } catch (final IOException e) {
      log.error("Problem in reading package ({})", this.archive, e);
    }
    return false;
  }

  public String getArchive() {
    return this.archive.toString();
  }

  public int getEntryNumber() {
    try (ZipFile zipFile = new ZipFile(this.getArchive())) {
      return zipFile.size();
    } catch (final IOException e) {
      log.error("IOException when getting entry number", e);
      return 0;
    }
  }

  public boolean unzipFiles(Path destination) {
    int count = 0;
    try (ZipInputStream zis = new ZipInputStream(
            new BufferedInputStream(new FileInputStream(this.getArchive()), SolidifyConstants.BUFFER_SIZE))) {
      ZipEntry ze = zis.getNextEntry();
      while (ze != null) {
        extractFile(zis, ze, destination, null);
        count++;
        ze = zis.getNextEntry();
      }
      zis.closeEntry();
    } catch (final IOException e) {
      log.error("Problem in unzipping package {} to destination {}", this.archive, destination, e);
      return false;
    }
    log.info("Unzip package ('{}') in '{}': {} entries extracted", this.archive, destination, count);
    return true;
  }

  public boolean unzipFiles(Path destination, List<String> list) {
    return this.unzipFiles(destination, list, null);
  }

  public boolean unzipFiles(URI destination) {
    return this.unzipFiles(Paths.get(destination));
  }

  public boolean unzipFolder(Path destination, String folder) {
    return this.unzipFiles(destination, Collections.singletonList(folder), folder);
  }

  public boolean zipFiles(Path source) {
    try {
      return this.createZip(source, this.generateFileList(source), null, false);
    } catch (final IOException e) {
      log.error("Problem in creating Zip package ({}.zip)", source, e);
      return false;
    }
  }

  public boolean zipFiles(URI source) {
    return this.zipFiles(Paths.get(source));
  }

  private boolean createZip(Path source, List<Path> relativePaths, List<Path> destinationPath, boolean excludeContainingFolder) {
    final byte[] buffer = new byte[SolidifyConstants.BUFFER_SIZE];
    int count = 0;
    try (ZipOutputStream zipOut = new ZipOutputStream(
            new BufferedOutputStream(new FileOutputStream(this.getArchive()), SolidifyConstants.BUFFER_SIZE))) {
      for (final Path file : relativePaths) {
        final Path f = source.resolve(file);
        ZipEntry zipEntry;
        if (destinationPath == null) {
          zipEntry = excludeContainingFolder ? new ZipEntry(normalizePath(file.getFileName().toString()))
                  : new ZipEntry(normalizePath(file.toString()));
        } else {
          zipEntry = excludeContainingFolder ? new ZipEntry(normalizePath(file.getFileName().toString()))
                  : new ZipEntry(normalizePath(destinationPath.get(count).toString()));
        }
        zipOut.putNextEntry(zipEntry);
        if (FileTool.isFile(f)) {
          try (BufferedInputStream bis = new BufferedInputStream(new FileInputStream(f.toFile()), SolidifyConstants.BUFFER_SIZE)) {
            int len;
            while ((len = bis.read(buffer)) > 0) {
              zipOut.write(buffer, 0, len);
            }
          }
        }
        count++;
        zipOut.closeEntry();
      }
    } catch (final FileNotFoundException e) {
      log.error("Cannot create Zip package ({})", this.archive, e);
      return false;
    } catch (final IOException e) {
      log.error("Problem in creating Zip package ({}.zip)", source, e);
      return false;
    }
    log.info("Zip package ('{}' in '{}') created with {} entries", source, this.archive, count);
    return true;
  }

  private List<Path> generateFileList(Path src) throws IOException {
    try (Stream<Path> walk = Files.walk(src)) {
      return walk.filter(path -> path.toFile().isFile()).map(src::relativize)
              .collect(Collectors.toList());
    }
  }

  private boolean unzipFiles(Path destination, List<String> list, String relativeFolder) {
    try (InputStream is = new BufferedInputStream(new FileInputStream(this.getArchive()), SolidifyConstants.BUFFER_SIZE)) {
      return unzipFiles(is, destination, list, relativeFolder);
    } catch (final IOException e) {
      log.error("Problem in unzipping package ({})", this.archive, e);
    }
    return false;
  }

}
